/*  $Header$
 *
 *  Copyright (C) 2003 Conexant Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#define __KERNEL_SYSCALLS__

/******************************************************************************
 *   UART over PCI Character driver
 ******************************************************************************/
#include <linux/version.h>
#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/init.h>
#include <linux/types.h>
#include <linux/unistd.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/pci.h>
#include <linux/string.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include "sm_drv_uartpci.h"
#include "sm_drv.h"
#include "sm_drv_pci.h"

static void uartpci_prot_exit(void);

extern struct net_device *root_sm_drv_device;

struct queue_s {
    uint32_t s;
    uint32_t p;
    uint32_t c;
    uint8_t *q;
};

typedef struct {
    uint32_t            uartpci_cts;            /* CTS pending */
    wait_queue_head_t   wait;                   /* read wait queue */
    spinlock_t          lock;
    uint32_t            major;

    struct queue_s      uartpci_txq;
    struct queue_s      uartpci_rxq;
} uartpci_priv_t;

/* The name for our device, as it will appear in /proc/devices */
#define DEVICE_NAME        "sm_uartpci"
#define UARTPCI_TXQSIZE     (32 * 1024)
#define UARTPCI_RXQSIZE     (32 * 1024)

static int uartpci_initq(struct queue_s *queue, int size);
static int uartpci_cleanupq(struct queue_s *queue);
static int uartpci_inq(struct queue_s *queue);
static uint8_t uartpci_qconsume(struct queue_s *queue);
static void uartpci_qproduce(struct queue_s *queue, uint8_t val);
static int uartpci_open(struct inode *inode, struct file *file);
static int uartpci_release(struct inode *inode, struct file *file);
static ssize_t uartpci_read(struct file *file, char *buffer, size_t length, loff_t *offset);
static ssize_t uartpci_write(struct file *file, const char *buffer, size_t length, loff_t *offset);

static int uartpci_is_open = 0;
static uartpci_priv_t uartpci_priv;
static uartpci_priv_t *upci = &uartpci_priv;

void uartpci_prot_init(void)
{
    const uint32_t uartpci_magic = cpu_to_le32(0x12345678);
    struct net_device *dev = root_sm_drv_device;
    struct net_local *lp = dev->priv;
    struct pci_hif_local_data   *hif_lp = ((struct pci_hif_local_data*)lp->hif_local_data);
    uint32_t reg;

    /* Disable CTS/DR interrupts */
    reg = readl(hif_lp->remapped_device_base + DEV38XX_INT_EN_REG);
    reg &= ~(DEV38XX_DEV_INT_PCIUART_CTS | DEV38XX_DEV_INT_PCIUART_DR);
    writel(reg, hif_lp->remapped_device_base + DEV38XX_INT_EN_REG);
    udelay(DEV38XX_WRITEIO_DELAY);

    /* write magic to activate UART via PCI protocol */
    writel(uartpci_magic, hif_lp->remapped_device_base + DEV38XX_CTRL_BLK_BASE_REG);
    mdelay(50);

    /* Check if the protocol is initialized */
    reg = readl(hif_lp->remapped_device_base + DEV38XX_INT_IDENT_REG);
    reg &= DEV38XX_DEV_INT_PCIUART_CTS;

    if(reg) {
        /* ACK the interrupt */
        writel(reg, hif_lp->remapped_device_base + DEV38XX_INT_ACK_REG);
        udelay(DEV38XX_WRITEIO_DELAY);
        DEBUG(DBG_IRQ, "UART PCI protocol initialized\n");

        /* Enable CTS/DR interrupts */
        reg = readl(hif_lp->remapped_device_base + DEV38XX_INT_EN_REG);
        reg |= (DEV38XX_DEV_INT_PCIUART_CTS | DEV38XX_DEV_INT_PCIUART_DR);
        writel(reg, hif_lp->remapped_device_base + DEV38XX_INT_EN_REG);
        udelay(DEV38XX_WRITEIO_DELAY);

        /* And give CTS to the device */
        writel(DEV38XX_DEV_INT_PCIUART_CTS, hif_lp->remapped_device_base + DEV38XX_DEV_INT_REG);
        udelay(DEV38XX_WRITEIO_DELAY);
    } else {
        DEBUG(DBG_IRQ, "UART PCI protocol not operational\n");
    }
}

void uartpci_prot_reinit(void)
{
    if (!uartpci_is_open)
        return;

    uartpci_prot_init();
}

static void uartpci_prot_exit(void)
{
    struct net_device *dev = root_sm_drv_device;
    struct net_local *lp = dev->priv;
    struct pci_hif_local_data   *hif_lp = ((struct pci_hif_local_data*)lp->hif_local_data);

    uint32_t reg;

    /* remove magic */
    writel(0, hif_lp->remapped_device_base + DEV38XX_CTRL_BLK_BASE_REG);
    udelay(DEV38XX_WRITEIO_DELAY);

    /* Disable CTS/DR interrupts */
    reg = readl(hif_lp->remapped_device_base + DEV38XX_INT_EN_REG);
    reg &= ~(DEV38XX_DEV_INT_PCIUART_CTS | DEV38XX_DEV_INT_PCIUART_DR);
    writel(reg, hif_lp->remapped_device_base + DEV38XX_INT_EN_REG);
    udelay(DEV38XX_WRITEIO_DELAY);
}


void uartpci_cts(struct net_device *dev)
{
    struct net_local *lp = dev->priv;
    struct pci_hif_local_data   *hif_lp = ((struct pci_hif_local_data*)lp->hif_local_data);
    uint32_t data;

    if(uartpci_inq(&upci->uartpci_txq) > 0) {
        data = cpu_to_le32(uartpci_qconsume(&upci->uartpci_txq));

        writel(data, hif_lp->remapped_device_base + DEV38XX_CTRL_BLK_BASE_REG);
        udelay(DEV38XX_WRITEIO_DELAY);

        /* Give DR */
        writel(DEV38XX_DEV_INT_PCIUART_DR, hif_lp->remapped_device_base + DEV38XX_DEV_INT_REG);
        udelay(DEV38XX_WRITEIO_DELAY);
    } else {
        /* Make sure magic is removed */
        writel(0, hif_lp->remapped_device_base + DEV38XX_CTRL_BLK_BASE_REG);
        udelay(DEV38XX_WRITEIO_DELAY);
        upci->uartpci_cts = 1;
    }
}

void uartpci_dr(struct net_device *dev)
{
    uint32_t reg;
    struct net_local *lp = dev->priv;
    struct pci_hif_local_data   *hif_lp = ((struct pci_hif_local_data*)lp->hif_local_data);

    reg = readl(hif_lp->remapped_device_base + DEV38XX_GP_COM_REG_2);

    if(uartpci_inq(&upci->uartpci_rxq) < UARTPCI_RXQSIZE) {
        uint8_t data = (uint8_t) (le32_to_cpu(reg) & 0xff);
        uartpci_qproduce(&upci->uartpci_rxq, data);
    }

    /* Give CTS */
    writel(DEV38XX_DEV_INT_PCIUART_CTS, hif_lp->remapped_device_base + DEV38XX_DEV_INT_REG);
    udelay(DEV38XX_WRITEIO_DELAY);

    /* Activate the bottom halve */
    wake_up_interruptible(&upci->wait);
}

static int uartpci_initq(struct queue_s *queue, int size)
{
    unsigned long flags;

    queue->q = kmalloc(size, GFP_ATOMIC);
    if(!queue->q) {
        printk(KERN_ERR "uartpci: Could not allocate queue\n");
        return -1;
    }

    spin_lock_irqsave(&upci->lock, flags);
    queue->p = queue->c = 0;
    queue->s = size;
    spin_unlock_irqrestore(&upci->lock, flags);

    return 0;
}

static int uartpci_cleanupq(struct queue_s *queue)
{
    if(queue->q)
        kfree(queue->q);

    return 0;
}

static int uartpci_inq(struct queue_s *queue)
{
    uint32_t inq;
    unsigned long flags;

    spin_lock_irqsave(&upci->lock, flags);
    inq = queue->p - queue->c;
    spin_unlock_irqrestore(&upci->lock, flags);

    return inq;
}

static uint8_t uartpci_qconsume(struct queue_s *queue)
{
    unsigned long flags;
    uint8_t ret;

    spin_lock_irqsave(&upci->lock, flags);
    ret = queue->q[queue->c % queue->s];
    (queue->c)++;
    spin_unlock_irqrestore(&upci->lock, flags);

    return ret;
}

static void uartpci_qproduce(struct queue_s *queue, uint8_t val)
{
    unsigned long flags;

    spin_lock_irqsave(&upci->lock, flags);
    queue->q[queue->p % queue->s] = val;
    (queue->p)++;
    spin_unlock_irqrestore(&upci->lock, flags);
}

static int uartpci_open(struct inode *inode, struct file *file)
{
    DEBUG(DBG_CALLS, "uartpci_open(%p,%p)\n", inode, file);

    /* We don't want to talk to two processes at the
     * same time */
    if (uartpci_is_open)
        return -EBUSY;

    uartpci_initq(&upci->uartpci_txq, UARTPCI_TXQSIZE);
    uartpci_initq(&upci->uartpci_rxq, UARTPCI_RXQSIZE);

    /* Initialize the UART over PCI debugging protocol */
    uartpci_prot_init();

    uartpci_is_open++;

    MOD_INC_USE_COUNT;

    return 0;
}

static int uartpci_release(struct inode *inode, struct file *file)
{
    DEBUG(DBG_CALLS, "uartpci_release(%p,%p)\n", inode, file);

    /* We're now ready for our next caller */
    uartpci_is_open--;

    uartpci_prot_exit();

    /* Decrement the usage count, otherwise once you
     * opened the file you'll never get rid of the module.
     */

    MOD_DEC_USE_COUNT;

    return 0;
}

static unsigned int uartpci_poll(struct file * filp, poll_table * wait)
{
    poll_wait(filp, &upci->wait, wait);

    if (uartpci_inq(&upci->uartpci_rxq) > 0)
            return POLLIN | POLLRDNORM;
    return 0;
}


static ssize_t uartpci_read(struct file *file,
                           char *buffer,    /* The buffer to fill with data */
                           size_t length,   /* The length of the buffer */
                           loff_t *offset)  /* Our offset in the file */
{
    int bytes_read;

    DEBUG(DBG_CALLS, "uartpci_read(%p,%p, %d)\n", file, buffer, length);

    if (uartpci_inq(&upci->uartpci_rxq) <= 0) {
        if (file->f_flags & O_NONBLOCK)
            return -EAGAIN;
        interruptible_sleep_on(&upci->wait);
        if (signal_pending(current))
            return -ERESTARTSYS;
    }

    bytes_read = 0;

    while(bytes_read < length &&
          bytes_read < uartpci_inq(&upci->uartpci_rxq)) {
        int ret;
        uint8_t data;

        data = uartpci_qconsume(&upci->uartpci_rxq);

        ret = put_user(data, buffer);
        if(ret < 0)
            return ret;

        buffer++;
        bytes_read++;
    }

    return bytes_read;
}


static ssize_t uartpci_write(struct file *file,
                            const char *buffer,    /* The buffer */
                            size_t length,   /* The length of the buffer */
                            loff_t *offset)  /* Our offset in the file */
{
    int written = 0;
    char c;
    int tries = 0;

    DEBUG(DBG_CALLS, "uartpci_write(%p,%p, %d)\n", file, buffer, length);

    while(written < length && tries < 10) {
        if(uartpci_inq(&upci->uartpci_txq) < UARTPCI_TXQSIZE) {
            if (get_user(c, buffer))
                return -EFAULT;

            uartpci_qproduce(&upci->uartpci_txq, c);
            written++;
            buffer++;
        } else if (current->need_resched) {
            schedule();
            tries++;
        }

        if(upci->uartpci_cts) {
            upci->uartpci_cts = 0;
            uartpci_cts(root_sm_drv_device);
        }
    }

    if(tries >= 10)
        printk(KERN_WARNING "UPCI: Write Q full, giving up\n");

    return written;
}


struct file_operations uartpci_fops = {
        read:           uartpci_read,
        write:          uartpci_write,
        open:           uartpci_open,
        release:        uartpci_release,
        poll:           uartpci_poll,
};


/* Initialize the module - Register the character device */
int uartpci_init_module(void)
{
    int ret;

    /* register our character device */
    if ((ret = register_chrdev(0, DEVICE_NAME, &uartpci_fops)) < 0) {
            printk(KERN_ERR "uartpci: can't get major number");
            return ret;
    }
    DEBUG(DBG_CALLS, "uartpci: registered dev major %d\n", ret);

    memset(upci, 0, sizeof(uartpci_priv_t));
    init_waitqueue_head(&upci->wait);
    spin_lock_init(&upci->lock);
    upci->major = ret;

    return 0;
}


/* Cleanup - unregister the appropriate file from /proc */
void uartpci_cleanup_module(void)
{
    int ret;

    DEBUG(DBG_CALLS, "uartpci_cleanup_module\n");

    uartpci_cleanupq(&upci->uartpci_txq);
    uartpci_cleanupq(&upci->uartpci_rxq);

    ret = unregister_chrdev(upci->major, DEVICE_NAME);

    if (ret < 0)
        printk(KERN_ERR "uartpci: error unregister_chrdev: %d\n", ret);
}

